# lop() and helper functions
#  started 18 Dec 2007, Walton Green, Gene Hunt, Scott Wing

############################################################
# lop is a function that simulates a tree, applies a ecological
#  character to the terminals with a given degree of phylogenetic
#  clumping and then subjects the tree to an extinction event.
lop <- function(n = 1, nt = 100, surv.size = nt/100,
                csize = nt/100, rt = FALSE, lambda = 0.6,
                xti = 0.7, dim = 1, xtt = 'rpoint',
                draw = TRUE, verbose = TRUE){
# n is number of different trees to synthesize
# nt is the number of terminals per tree
# surv.size is a number between 1 and nt defaulting to nt/100 
#  recording the number of surviving clades with size greater
#  than or equal to surv.size.
# csize is the size clade at which to count proportion of extinct
#  clades; i.e. for csize = 2, the csizes.p return array is the
#  number of clades of size 2 that go extinct and csizes.t is the
#  number of clades of size 2 total
# rt is a logical indicating whether the csize measurement
#  is two-sided or just right tail. Default (rt = FALSE) gives
#  two-sided output.
# lambda is lambda
# xti is extinction proportion
# dim is the number of dimensions of ecological variability
# xtt is the type of extinction: 'center' hits the center
#  of the ecological variable; 'top' hits the high end;
#  'rpoint' picks a terminal at random on which to center the
#  extinction; 'rand' kills entirely at random with respect
#  location in ecological space
# draw is logical indicating whether to produce graphic output

# nt, lambda, and xti can all be vectors

# install the following packages, if necessary:
ape.loaded <- require(ape)
mass.loaded <- require(MASS)
geiger.loaded <- require(geiger)
if(!(ape.loaded || mass.loaded || geiger.loaded))
  stop('need to install packages ape, MASS, and geiger')

start.time <- date()

############################################################
## main

 # this is an array of the right shape to hold collected data
 data <- array(dim = c(n, length(lambda), length(xti)),
               dimnames = list(
                 n = as.character(1:n),
                 lambda = paste('lambda=', lambda, sep = ''),
                 xti = paste('xti=', xti, sep = '')
                 )
               )
               
 # instances of this data array
 largest.ext.sizes <- data
 mean.ext.sizes <- data
 ext.gt.surv.sizes <- data
 csizes.p <- data
 csizes.t <- data
 ext.tables <- array(list(), dim = c(n, length(lambda),
                     length(xti)),
                     dimnames = list(
                       n = as.character(1:n),
                       lambda = paste('lambda=', lambda,
                                      sep = ''),
                       xti = paste('xti=', xti, sep = '')
                 )
               )

 this.tree <- 0
 for (i in 1:n){ #loop through trees
  # generate tree with ecologial traits
  #tr <- rcoal(nt)
  # this was the old way of creating a tree from ape;
  # not clear that coalescent tree is what we want,
  # so switched to a birth-death model
  tr <- birthdeath.tree(b = 1, d = 0, taxa.stop = nt + 1)
  tr <- drop.tip(tr, tip = as.character(nt + 1))
  tr <- read.tree(text = write.tree(tr, file = ""))
  tr$tip.label <- paste("t", tr$tip.label, sep = "")
  VV <- vcv.phylo(tr)
  for (j in 1:length(lambda)){ #loop through values of lambda
   # lambda
   Lambda <- matrix(data = lambda[j], nrow = nt, ncol = nt)
   diag(Lambda) <- 1
   
   #create an ecological variable across dim dimensions
   ecoV <- matrix(0, nrow = dim, ncol = nt)
   for (ii in 1:dim){
     ecoV[ii,] <- mvrnorm(n = 1, mu = rep(0, times = nt),
                     Sigma = Lambda * VV)
   }
   colnames(ecoV) <- tr$tip.label
   
   for (k in 1:length(xti)){ #loop through values of xti
   	this.tree <- this.tree + 1
    # determine extinctions based on ecoV
    ne <- round(nt * xti[k]) #ne is number of terminals killed
    
    if(dim == 1){ # in the 1-dimensional case
      ecoV <- as.vector(ecoV)
      names(ecoV) <- tr$tip.label     
       if(xtt == 'center'){
        s.devs <- sort(abs(ecoV))
        dead.taxa <- names(s.devs[1:ne])
        live.index <- !(names(ecoV) %in% dead.taxa)
      }else if(xtt == 'top'){
        s.devs <- sort(ecoV)
        dead.taxa <- names(s.devs[1:ne])
        live.index <- !names(ecoV) %in% dead.taxa    
      }else if(xtt == 'rand'){
        s.devs <- ecoV[sample(length(ecoV))]
        dead.taxa <- names(s.devs[1:ne])
        live.index <- !names(ecoV) %in% dead.taxa
      }else if(xtt == 'rpoint'){
      	# chooses random taxon to center extinction on
      	rp<- sample(1:length(ecoV), size=1)
      	ecoV<- ecoV - ecoV[rp]
      	# centers ecoV on random point, then same as xtt='center'
        s.devs <- sort(abs(ecoV))
        dead.taxa <- names(s.devs[1:ne])
        live.index <- !(names(ecoV) %in% dead.taxa)             
      }else{
        stop('argument to xtt not understood')
      }
      if (draw) {
      	plot(ecoV, pch=19, col=2-live.index)
      	abline(h=0, lty=3)
        system('sleep 10')}
    }else{ # for >1-dimensional case
      if(xtt == 'center'){
        distfromcenter <- function(x){
          meanloc <- apply(ecoV, 1, mean)
      	  as.numeric(dist(rbind(x,meanloc)))
        }
        s.devs <- sort(apply(ecoV, 2, distfromcenter))
        dead.taxa <- names(s.devs[1:ne])
        live.index <- !colnames(ecoV) %in% dead.taxa
      }else if(xtt == 'top'){
        distfromtop <- function(x){
          top <- apply(ecoV, 1, max)
      	  as.numeric(dist(rbind(x,top)))
        }
        s.devs <- sort(apply(ecoV, 2, distfromtop))
        dead.taxa <- names(s.devs[1:ne])
        live.index <- !colnames(ecoV) %in% dead.taxa    
      }else if(xtt == 'rand'){
        s.devs <- ecoV[sample(nt)]
        dead.taxa <- names(s.devs[1:ne])
        live.index <- !colnames(ecoV) %in% dead.taxa
      }else if(xtt == 'rpoint'){
      	# chooses random taxon to center extinction on
      	rp<- sample(1:nt, size=1)
      	ecoV<- t(scale(t(ecoV), center = ecoV[,rp], scale=FALSE)) 
      	# centers ecoV on random point
		ff<- function(x) sqrt(sum(x^2))
		d2rp<- apply(ecoV, 2, FUN=ff)  # dist to random point
        s.devs <- sort(d2rp)
        dead.taxa <- names(s.devs[1:ne])
        live.index <- !colnames(ecoV) %in% dead.taxa
      }else{
        stop('argument to xtt not understood')
      }
      if(draw){
      	ff <- function(x, y, ...){
      	  points(x,y, ...)
      	  abline(h=0, lty=3)
      	  abline(v=0, lty=3)
      	}
      	plot(as.data.frame(t(ecoV)),
      	     col = 2 - live.index)
        system('sleep 10')
      }
    }
    
    # measures of change
    #all.sons <- get.all.sons(tr)
    dcs <- dead.clade.sizes(tr, dead.tips = dead.taxa)
    largest.ext.size <- max(dcs$dead.distr)
    largest.ext.sizes[i,j,k] <- largest.ext.size
    mean.ext.size <- mean(dcs$dead.distr)
    mean.ext.sizes[i,j,k] <- mean.ext.size
    ext.gt.surv.size <- sum(dcs$dead.distr > surv.size)
    ext.gt.surv.sizes[i,j,k] <- ext.gt.surv.size
    if(rt){
      csize.p <- sum(dcs$sizes[dcs$ext] >= csize)
      csize.t <- sum(dcs$sizes >= csize)
    }else{
      csize.p <- sum(dcs$sizes[dcs$ext] == csize)
      csize.t <- sum(dcs$sizes == csize)
    }
    csizes.p[i,j,k] <- csize.p
    csizes.t[i,j,k] <- csize.t

    ext.tables[[i,j,k]] <- table(dcs$dead.distr)

    if(verbose){
      cat(paste('Simulating tree', this.tree ,'of',
                length(data) ,'with', nt, 'terminals...\n'))
      cat(paste('Extinction intensity: ', xti[k], '\n', sep = ''))
      cat(paste('Phylogenetic clumping (lambda): ',
                lambda[j], '\n', sep = ''))
      cat('Table of clade sizes:')
      print(table(dcs$sizes))
      cat('Table of extinct clade sizes:')
      print(table(dcs$dead.distr))
      cat(paste('Mean size of extinct clades: ',
                format(mean.ext.size, digits = 4),
                '\n', sep = ''))
      cat(paste('Size of largest extinct clade: ',
                largest.ext.size, '\n', sep = ''))
    }
    
    if(draw){
     edge.color <- rep('black', length(dcs$nodes))
     draw.tree(tr, ecoV, index = live.index, n = i,
               lambda = lambda[j], xti = xti[k],
               edge.color = edge.color)
     system('sleep 1')
    }
    cat('==============================\n')
   } # end loop through xti extinction intenstities
  } # end loop through lambda values
 } # end loop through trees
 
 end.time <- date()
 if(verbose) cat(paste('Started:', start.time, 'Finished:',
                        end.time, '\n'))
 
 return(list(param = c(n = n, nt = nt, surv.size = surv.size,
                       csize = csize, rt = rt, dim = dim,
                       lambda = lambda, xti = xti, xtt = xtt),
             largest.ext.sizes = largest.ext.sizes,
             mean.ext.sizes = mean.ext.sizes,
             ext.gt.surv.sizes = ext.gt.surv.sizes,
             csizes.p = csizes.p, csizes.t = csizes.t,
             ext.tables = ext.tables,
             times = list(start = start.time,
                          end = end.time)))
} # end of lop function

############################################################
# sub
scale1to5 <- function(x){
 xm <- x - min(x)
 xs <- 0.5 + 4.5 * xm/max(xm)
 return(xs)
}

############################################################
# see plot.phylo()
draw.tree <- function(tr, ecoV, index = NULL, n, lambda,
                      xti, edge.color){
 if(is.matrix(ecoV)) ecoV <- apply(ecoV, 2, mean)
 plot(tr, show.tip.label = FALSE, show.node.label = FALSE,
      edge.color = edge.color, label.offset = 0.1,
      tip.color = edge.color)
 mtext(paste('tree:', n, ' lambda:', lambda, ' xti:',
             xti, sep = ''))
 if (is.null(index))  bgs <- rep(3, times=length(ecoV))
 else  bgs <- 2 - as.numeric(index)
 #tiplabels(pch = 21, bg = bgs, cex = scale1to5(ecoV), col="white")
 tiplabels(pch = 21, bg = bgs)
}

############################################################
## function to (ultimately) return distribution 
## extinct clade sizes

get.all.sons<- function(tr)
# returns list of all sons for each node & tip
{
 ee<- tr$edge
 nT<- length(tr$tip.label)
 nN<- tr$Nnode
 sons<- list()
 
 for (i in 1:(nT+nN))
   sons[[i]]<- node.sons(tr,i)	
 
 return(sons)
}

node.sons<- function(tr, node)
# returns vector of sons for a particular node
{
 ee<- tr$edge
 nT<- length(tr$tip.label)
 nN<- tr$Nnode
 d.tips<- numeric()
 d.nodes<- numeric()
    
 desc<- ee[ee[,1]==node,2]
 #print (desc)
 d.tips<- append(d.tips, desc[desc<=nT])
 d.nodes<- append(d.nodes, desc[desc>nT])
 
 if (node<=nT)	d.tips<- node
 else{
 while(!all(desc<=nT))  # true if some desc are nodes not tips
  {
  desc<- ee[ee[,1] %in% desc,2]
  #print(desc)
  d.tips<- append(d.tips, desc[desc<=nT])
  d.nodes<- append(d.nodes, desc[desc>nT])  	
  }
 }
 return(list(tips=d.tips, nodes=d.nodes, node.label=node))
}

dead.clade.sizes<- function(tr, dead.tips)
# return list of all tips descended from each node
# dead.tips are the labels of dead terminal taxa
{
 # get all clades
 clades<- get.all.sons(tr)
 nc<- length(clades)

 # get sizes of clades
 ff1<- function (x) {length(x$tips)}
 clade.sizes<- sapply(clades, FUN=ff1)
 
 # which clades are extinct
 dd<- match(dead.tips, tr$tip.label)
 ff2<- function (x){all(x$tips %in% dd)}
 ext<- sapply(clades, FUN=ff2)
  
 # retain largest of nested extinct clades 
 ext.keep<- ext
 wh<- which(ext)
 for (i in wh)
  {
    for (j in wh)	
  	 {
  	   ok1<- i!=j 
  	   ok2<- clades[[i]]$node.label %in% c(clades[[j]]$tips, clades[[j]]$nodes)
  	   ok3<- clade.sizes[i] < clade.sizes[j]
  	   ok4<- ext[j]
  	   if (ok1 & ok2 & ok3 & ok4)	ext.keep[i]<- FALSE    
  	   #cat (i,j,ok1, ok2, ok3, "\n", sep="  ")	
  	 	}
  }   
 w<- list(nodes=1:length(clades), ext=ext, ext.keep=ext.keep, sizes=clade.sizes, dead.distr=clade.sizes[ext.keep])
 return(w)	
}
#$node: these are the nodes numbers, taken from the tree's
# edge matrix
#$ext: logical, indicating if the relevant node is extinct at
# the end
#$ext.keep: logical, indicating if a node is extinct AND the
# largest of nested extinct nodes
#$sizes: the sizes (# tips descended from each node)
#$dead.distr: the distribution of the sizes of extinct clades.
############################################################